# (c) Copyright IBM Corp. 2010, 2017. All Rights Reserved.

import pkg_resources
try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:
    pass

from .misc import verify_subset
from .mocks import BasicResilientMock, resilient_endpoint
from .circuits_fixtures import call_event, call_event_from_name, wait_for



# (c) Copyright IBM Corp. 2010, 2017. All Rights Reserved.

"""Pytest Fixture for Testing Resilient Circuits Components"""
from __future__ import print_function

import os
import sys
import time
import calendar
import json
import traceback
import logging
import pytest
from pytest_resilient_circuits.circuits_fixtures import manager, watcher
import resilient
from resilient import SimpleHTTPException
import resilient_circuits.app

DATATABLE_TYPE_ID = 8


class ConfiguredAppliance:
    """ configure resilient org with specs from the test module """
    def __init__(self, request):
        # TODO: replace/supplement with the 'resilient.circuits.customize' entry-point
        # TODO: Add support for phases, tasks, incident types.
        # TODO: Add support for optional action and custom fields (currently all set as required)

        host = os.environ.get("TEST_RESILIENT_APPLIANCE", request.config.option.resilient_host)
        url = "https://%s:443" % host
        org = os.environ.get("TEST_RESILIENT_ORG", request.config.option.resilient_org)
        user = os.environ.get("TEST_RESILIENT_USER", request.config.option.resilient_email)
        password = os.environ.get("TEST_RESILIENT_PASSWORD", request.config.option.resilient_password)
        assert all((host, org, user, password))

        # Connect to Resilient
        self.client = resilient.SimpleClient(org_name=org, base_url=url, verify=False)
        session = self.client.connect(user, password)

        # Retrieve constants from appliance
        constants = self._get_constants()["actions_framework"]
        # Invert the dictionaries so they are {"name": id,...}
        action_object_types = {value: int(key) for key, value in constants["action_object_types"].items()}
        action_types = {value: int(key) for key, value in constants["action_types"].items()}
        destination_types = {value: int(key) for key, value in constants["destination_types"].items()}

        # Delete all existing configuration items
        self._clear_org()

        # Create all required configuration items for this class of tests
        action_fields = getattr(request.cls, "action_fields", None)
        if action_fields:
            for field_name, info in action_fields.items():
                success = self._create_action_field(field_name,
                                                    info[0], info[1], info[2])
                assert success
        custom_fields = getattr(request.cls, "custom_fields", None)
        if custom_fields:
            for field_name, info in custom_fields.items():
                success = self._create_custom_field(field_name,
                                                    info[0], info[1], info[2])
                assert success
        destinations = getattr(request.cls, "destinations", None)
        if destinations:
            for queue_name in destinations:
                success = self._create_destination(queue_name,
                                                   destination_types["Queue"])
                assert success

        destinations = self.client.get("/message_destinations")["entities"]
        destinations = {dest["programmatic_name"]: int(dest["id"]) for dest in destinations}
        manual_actions = getattr(request.cls, "manual_actions", None)
        if manual_actions:
            for action_name, info in manual_actions.items():
                success = self._create_manual_action(action_name,
                                                     destinations[info[0]],
                                                     action_object_types[info[1]],
                                                     action_types["Manual"],
                                                     info[2])
                assert success
        automatic_actions = getattr(request.cls, "automatic_actions", None)
        if automatic_actions:
            for action_name, info in automatic_actions.items():
                success = self._create_automatic_action(action_name,
                                                        destinations[info[0]],
                                                        action_object_types[info[1]],
                                                        action_types["Automatic"],
                                                        info[2])
                assert success

        data_tables = getattr(request.cls, "data_tables", None)
        print("create data tables: %s" % data_tables)
        if data_tables:
            for table_name, columns in data_tables.items():
                success = self._create_data_table(table_name, columns)
                assert success

    # end __init__

    def _clear_org(self):
        """ Delete all existing destinations, actions, fields, etc from org """

        actions = self.client.get("/actions")["entities"]
        if actions:
            for action in actions:
                self.client.delete("/actions/%s" % action['id'])

        destinations = self.client.get("/message_destinations")["entities"]
        if destinations:
            for destination in destinations:
                print("/message_destinations/%s" % destination['id'])
                self.client.delete("/message_destinations/%s" % destination['id'])

        fields = self.client.get("/types/actioninvocation/fields")
        if fields:
            for field in fields:
                self.client.delete("/types/actioninvocation/fields/%s" % field['id'])

        fields = self.client.get("/types/incident/fields")
        if fields:
            fields = [field for field in fields if field['prefix'] == "properties"]
            for field in fields:
                print("URL IS /types/incident/fields/%s" % field['id'])
                self.client.delete("/types/incident/fields/%s" % field['id'])

        types = self.client.get("/types")
        data_tables = [res_type["type_name"] for res_type in types.values() if res_type['type_id'] == DATATABLE_TYPE_ID]
        for dt_name in data_tables:
            print("Delete /types/%s" % dt_name)
            self.client.delete("/types/%s" % dt_name)
    # end _clear_org

    def _get_constants(self):
        """ get data from /rest/const """
        url = "{0}/rest/const".format(self.client.base_url)
        response = self.client._execute_request(self.client.session.get,
                                                url,
                                                proxies=self.client.proxies,
                                                cookies=self.client.cookies,
                                                headers=self.client._SimpleClient__make_headers())
        if response.status_code != 200:
            raise SimpleHTTPException(response)
        return json.loads(response.text)
    # end _get_constants

    def _create_action_field(self, programmatic_name, field_type, display_name, values):
        """ Create action field in resilient """
        endpoint = "/types/actioninvocation/fields"
        action_field = {"text": display_name,
                        "required": "always",
                        "blank_option": False,
                        "input_type": field_type,
                        "name": programmatic_name}
        if values:
            action_field["values"] = [{"label": value} for value in values]

        try:
            field_def = self.client.post(endpoint, action_field)
            if not field_def:
                print("Failed to create action field %s" % programmatic_name)
                return False
        except Exception as e:
            print("Failed to create action field %s" % programmatic_name)
            traceback.print_exc()
            return False
        return True
    # end _create_action_field

    def _create_custom_field(self, programmatic_name, field_type, display_name, values):
        """ Create custom field in resilient """
        endpoint = "/types/incident/fields"
        custom_field = {"text": display_name,
                        "input_type": field_type,
                        "name": programmatic_name}
        if values:
            custom_field["values"] = [{"label": value} for value in values]

        response = self.client.post(endpoint, custom_field)
        return True if response else False
    # end _create_custom_field

    def _create_destination(self, name, destination_type):
        """ Create destination queue """
        user_id = self.client.user_id
        endpoint = "/message_destinations"
        destination_obj = {"name": name,
                           "expect_ack": True,
                           "destination_type": destination_type,
                           "programmatic_name": name,
                           "users": [user_id]}
        try:
            destination = self.client.post(endpoint, destination_obj)
            if not destination:
                print("Failed to create destination %s" % name)
                return False
        except Exception as e:
            print("Failed to create destination %s" % name)
            traceback.print_exc()
            return False
        return True
    # end _create_destination

    def _create_manual_action(self, action_name, destination_id, object_type_id, action_type_id, fields):
        """ Create and configure a manual action """
        endpoint = "/actions"
        action = {"name": action_name,
                  "type": action_type_id,
                  "object_type": object_type_id,
                  "message_destinations": [destination_id],
                  "view_items": [{"field_type": "actioninvocation",
                                  "element": "field",
                                  "content": fieldname} for fieldname in fields]
                  }
        try:
            action_obj = self.client.post(endpoint, action)
            if not action_obj:
                print("Failed to create action %s" % action_name)
                return False
        except Exception as e:
            print("Failed to create action %s" % action_name)
            traceback.print_exc()
            return False

        return True
    # end _create_manual_action

    def _create_automatic_action(self, action_name, destination_id, object_type_id, action_type_id, conditions):
        """ Create and configure automatic action """
        endpoint = "/actions"
        action = {"name": action_name,
                  "type": action_type_id,
                  "object_type": object_type_id,
                  "message_destinations": [destination_id]
                  }
        if conditions:
            action["conditions"] = conditions
        try:
            action_obj = self.client.post(endpoint, action)
            if not action_obj:
                print("Failed to create action %s" % action_name)
                return False
        except Exception as e:
            print("Failed to create action %s" % action_name)
            traceback.print_exc()
            return False

        return True
    # end _create_automatic_action

    def _create_data_table(self, table_name, columns):
        """" Create a data table """
        endpoint = "/types"
        table_columns = {}
        print("create table %s" % table_name)
        table = {"type_name": table_name,
                 "display_name": table_name,
                 "type_id": DATATABLE_TYPE_ID,
                 "parent_types": ["incident", ]}
        try:
            dt_obj = self.client.post(endpoint, table)
            if not dt_obj:
                print("Failed to create data table %s" % table_name)
                return False
        except Exception as e:
            print("Failed to create data table %s" % table_name)
            traceback.print_exc()
            return False

        endpoint = "/types/%s/fields" % table_name
        for col_name, (col_type, col_values) in columns.items():
            if not col_values:
                values = []
            else:
                values = [{"label": value} for value in col_values]
            field = {"name": col_name,
                     "text": col_name,
                     "input_type": col_type,
                     "values": values}
            try:
                field_obj = self.client.post(endpoint, field)
                if not field_obj:
                    print("Failed to create data table field %s" % col_name)
                    return False
            except Exception as e:
                print("Failed to create data table field %s" % col_name)
                traceback.print_exc()
                return False

        return True
    # end _create_data_table

# end ConfiguredAppliance


class ResilientCircuits:
    def __init__(self, tmpdir_factory, manager, watcher, request):

        # Reset the rest client b/c it is a global variable
        resilient_circuits.rest_helper.reset_resilient_client()
        resilient_config_data = """
[resilient]
logfile = app.log
loglevel = INFO
cafile = false
stomp_port = 65001
no_prompt_password = True
port = 443
test_actions = True
"""
        print("CURRENT WORKING DIR:  Addr: ", os.getcwd(), id(self))

        resilient_mock = getattr(request.module, "resilient_mock", None)
        self.config_file = tmpdir_factory.mktemp('data').join("%dapp.config" % id(self))
        print("TEST config={}".format(self.config_file.strpath))

        self.logs = tmpdir_factory.mktemp("logs")
        print("TEST logdir={}".format(self.logs))

        config_data = getattr(request.module, "config_data", "")
        self.config_file.write(resilient_config_data)

        host = os.environ.get("TEST_RESILIENT_APPLIANCE")
        if host:
            print("TEST host={} (from environment)".format(host))
        else:
            host = request.config.option.resilient_host
            print("TEST host={} (from configuration data)".format(host))
        self.config_file.write("host = %s\n" % host, mode='a')

        self.org = os.environ.get("TEST_RESILIENT_ORG")
        if self.org:
            print("TEST org={} (from environment)".format(self.org))
        else:
            self.org = request.config.option.resilient_org
            print("TEST org={} (from configuration data)".format(self.org))
        self.config_file.write("org = %s\n" % self.org, mode='a')

        self.user = os.environ.get("TEST_RESILIENT_USER")
        if self.user:
            print("TEST user={} (from environment)".format(self.user))
        else:
            self.user = request.config.option.resilient_email
            print("TEST user={} (from configuration data)".format(self.user))
        self.config_file.write("email = %s\n" % self.user, mode='a')

        password = os.environ.get("TEST_RESILIENT_PASSWORD", request.config.option.resilient_password)
        self.config_file.write("password = %s\n" % password, mode='a')
        self.config_file.write("log_http_responses = %s\n" % self.logs, mode='a')
        self.config_file.write("logdir = %s\n" % self.logs, mode='a')
        self.config_file.write("test_port = 0\n", mode='a')
        if resilient_mock:
            if isinstance(resilient_mock, type):
                self.config_file.write("resilient_mock = %s.%s\n" % (resilient_mock.__module__, resilient_mock.__name__), mode='a')
            else:
                self.config_file.write("resilient_mock = %s\n" % resilient_mock, mode='a')

        self.config_file.write(config_data, mode='a')
        os.environ["APP_CONFIG_FILE"] = self.config_file.strpath
        # Set this manually b/c it is only read on import in app.py
        resilient_circuits.app.APP_CONFIG_FILE = self.config_file.strpath

        self.manager = manager
        self.watcher = watcher

        # Remove the pytest commandline arguments so they don't break ArgParse in resilient
        sys.argv = sys.argv[0:1]

        self.app = resilient_circuits.app.App().register(manager)
        assert watcher.wait("registered")
        pytest.wait_for(manager, "_running", True)
        assert watcher.wait("load_all_success", timeout=10)

    # end __init__

    def finalizer(self):
        """ Unregister and reset CWD back to where it was originally """
        print("Finalizer Running")
        self.flush_logs()
        self.manager.stop()
        pytest.wait_for(self.manager, "_running", False)

    def flush_logs(self):
        handlers = logging.getLogger().handlers
        for handler in handlers:
            if isinstance(handler, logging.FileHandler) or isinstance(handler, logging.handlers.RotatingFileHandler):
                handler.flush()

# end ResilientCircuits


@pytest.fixture(scope="class")
def circuits_app(tmpdir_factory, manager, watcher, request):
    circuits_app_fixture = ResilientCircuits(tmpdir_factory, manager, watcher, request)
    yield circuits_app_fixture
    circuits_app_fixture.finalizer()


@pytest.fixture(scope="class")
def configure_resilient(request):
    """ Create necessary destinations, actions, fields, etc in Resilient """
    yield ConfiguredAppliance(request)
# end configure_resilient


@pytest.fixture(scope="class")
def new_incident(circuits_app):
    """ Create but don't post a minimally populated incident for testing """
    client = circuits_app.app.action_component.rest_client()
    incident = {}
    fields = client.get('/types/incident/fields')
    for field_def in fields:
        if field_def.get('required', '') != 'always':
            continue
        fieldname = field_def['name']
        field_type = field_def['input_type']

        if field_def['values']:
            # Just use whatever the first valid value is
            value = field_def['values'][0]["value"]
        elif field_type == "boolean":
            value = True
        elif field_type in ("text", "textarea"):
            value = "test"
        elif field_type == "datetimepicker":
            value = int(calendar.timegm(time.gmtime())) * 1000
        elif field_type == "number":
            value = 1

        if field_def['prefix']:
            incident[field_def['prefix']][fieldname] = value
        else:
            incident[fieldname] = value
    yield incident
# end new_incident



# (c) Copyright IBM Corp. 2010, 2017. All Rights Reserved.
"""py.test config"""
from __future__ import print_function

import pytest

import sys
import threading
import collections
from time import sleep
from collections import deque

from circuits.core.manager import TIMEOUT
from circuits import handler, BaseComponent, Debugger, Manager


class Watcher(BaseComponent):

    def init(self):
        self._lock = threading.Lock()
        self._events = deque()

    @handler(channel="*", priority=999.9)
    def _on_event(self, event, *args, **kwargs):
        # print("WATCHER GOT ", event)
        with self._lock:
            self._events.append(event)

    def clear(self):
        self._events.clear()

    def wait(self, name, parent=None, channel=None, timeout=6.0):
        for i in range(int(timeout / TIMEOUT)):
            with self._lock:
                for event in self._events:
                    if event.name == name and event.waitingHandlers == 0:
                        if (channel is None) or (channel in event.channels):
                            if parent:
                                # match a parent of this event
                                p = event
                                while p:
                                    if p == parent:
                                        return event
                                    p = p.parent
                            else:
                                e = event
                                self._events.remove(event)
                                return e
            sleep(TIMEOUT)
        else:
            return False


class Flag(object):
    status = False


def call_event_from_name(manager, event, event_name, *channels):
    fired = False
    value = None
    for r in manager.waitEvent(event_name):
        if not fired:
            fired = True
            value = manager.fire(event, *channels)
        sleep(0.1)
    return value


def call_event(manager, event, *channels):
    return call_event_from_name(manager, event, event.name, *channels)


class WaitEvent(object):

    def __init__(self, manager, name, channel=None, timeout=6.0):
        if channel is None:
            channel = getattr(manager, "channel", None)

        self.timeout = timeout
        self.manager = manager

        flag = Flag()

        @handler(name, channel=channel)
        def on_event(self, *args, **kwargs):
            flag.status = True

        self.handler = self.manager.addHandler(on_event)
        self.flag = flag

    def wait(self):
        try:
            for i in range(int(self.timeout / TIMEOUT)):
                if self.flag.status:
                    return True
                sleep(TIMEOUT)
        finally:
            self.manager.removeHandler(self.handler)


def wait_for(obj, attr, value=True, timeout=3.0):
    from circuits.core.manager import TIMEOUT
    for i in range(int(timeout / TIMEOUT)):
        if isinstance(value, collections.Callable):
            if value(obj, attr):
                return True
        elif getattr(obj, attr) == value:
            return True
        sleep(TIMEOUT)


@pytest.fixture(scope="class")
def manager(request):
    manager = Manager()

    def finalizer():
        manager.stop()

    request.addfinalizer(finalizer)

    waiter = WaitEvent(manager, "started")
    manager.start()
    assert waiter.wait()

    if request.config.option.verbose:
        verbose = True
    else:
        verbose = False

    Debugger(events=verbose).register(manager)

    return manager


@pytest.fixture(scope="class")
def watcher(request, manager):
    watcher = Watcher().register(manager)

    def finalizer():
        waiter = WaitEvent(manager, "unregistered")
        watcher.unregister()
        waiter.wait()

    request.addfinalizer(finalizer)

    return watcher


def pytest_namespace():
    return dict((
        ("WaitEvent", WaitEvent),
        ("wait_for", wait_for),
        ("call_event", call_event),
        ("PLATFORM", sys.platform),
        ("PYVER", sys.version_info[:3]),
        ("call_event_from_name", call_event_from_name),
    ))



""" Test utilities for Resilient """


def verify_subset(expected, actual):
    """Test that the values match, where expected can be a subset of actual"""
    if isinstance(expected, dict):
        assert isinstance(actual, dict)
        for (key, value) in expected.items():
            verify_subset(value, actual.get(key))
    elif isinstance(expected, list):
        assert isinstance(actual, list)
        for evalue, avalue in zip(expected, actual):
            verify_subset(evalue, avalue)
    else:
        assert expected == actual



# (c) Copyright IBM Corp. 2010, 2017. All Rights Reserved.

""" py.test plugin configuration for pytest-resilient-circuits """

from pytest_resilient_circuits.circuits_fixtures import *
from pytest_resilient_circuits.resilient_circuits_fixtures import *


_help_email = "Resilient user email"
_help_host = "Resilient host"
_help_password = "Resilient password"
_help_org = "Resilient org"


def pytest_addoption(parser):
    """ Configuration Options """
    parser.addoption("--resilient_email",
                     help=_help_email,
                     action="store",
                     default="",
                     required=False)

    parser.addoption("--resilient_host",
                     help=_help_host,
                     action="store",
                     default="",
                     required=False)

    parser.addoption("--resilient_password",
                     help=_help_password,
                     action="store",
                     default="",
                     required=False)

    parser.addoption("--resilient_org",
                     help=_help_org,
                     action="store",
                     default="",
                     required=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2017. All Rights Reserved.

""" Setup for pytest_resilient_circuits module """

import os
from setuptools import setup

requires_resilient_version = "29.0"
major, minor = requires_resilient_version.split('.', 2)[:2]

requirements = [
    'pytest>=3.0.0',
    'resilient>={}.{}'.format(major, minor),
    'resilient-circuits>={}.{}'.format(major, minor)
]

setup(
    name='pytest_resilient_circuits',
    use_scm_version={"root": "../", "relative_to": __file__},
    setup_requires=['setuptools_scm'],
    url='https://developer.ibm.com/resilient',
    license='MIT',
    author='IBM Resilient',
    author_email='support@resilientsystems.com',
    description='Resilient Circuits fixtures for PyTest.',
    packages=['pytest_resilient_circuits'],
    package_data={'pytest_resilient_circuits': ['version.txt']},
    install_requires=requirements,
    include_package_data=True,
    entry_points={
        'pytest11': [
            'pytest_resilient_circuits = pytest_resilient_circuits.plugin'
        ]},
)



""" Requests mock for Resilient REST API """

import logging
import json
import io
import requests
import requests_mock
import pkg_resources
from resilient.resilient_rest_mock import ResilientMock, resilient_endpoint
LOG = logging.getLogger(__name__)


def test_data(filename):
    """Read a test JSON data file"""
    template_file_path = pkg_resources.resource_filename("pytest_resilient_circuits", "data/{}".format(filename))
    with io.open(template_file_path, 'r', encoding="utf-8") as template_file:
        return json.load(template_file)


class BasicResilientMock(ResilientMock):
    """Mock that implements some of the basic REST endpoints.  Test classes can extend this where needed."""

    def __init__(self, *args, **kwargs):
        super(BasicResilientMock, self).__init__(*args, **kwargs)
        self.incident = test_data("200_JSON_GET__incidents_2314.json")

    @resilient_endpoint("POST", "/rest/session")
    def session_post(self, request):
        """ Callback for POST to /rest/session """
        LOG.debug("session_post")
        session_data = {
            "saml_alias": None,
            "csrf_token": "79945884c2e6f2339cbffbbaba01f17b",
            "user_lname": "Doe",
            "user_id": 1,
            "is_ldap": False,
            "is_saml": False,
            "orgs": [
                {
                    "city": "AnyCity",
                    "addr": None,
                    "zip": None,
                    "has_available_twofactor": False,
                    "perms": {
                        "create_shared_layout": True,
                        "administrator": False,
                        "create_incs": True,
                        "master_administrator": True,
                        "observer": False
                    },
                    "supports_ldap": False,
                    "enabled": True,
                    "twofactor_auth_domain": None,
                    "attachments_enabled": True,
                    "has_saml": False,
                    "state": None,
                    "addr2": None,
                    "twofactor_cookie_lifetime_secs": 0,
                    "require_saml": False,
                    "tasks_private": False,
                    "authorized_ldap_group": None,
                    "id": 201,
                    "name": self.org_name
                },
                {
                    "city": None,
                    "addr": None,
                    "zip": None,
                    "has_available_twofactor": False,
                    "perms": {
                        "create_shared_layout": True,
                        "administrator": False,
                        "create_incs": True,
                        "master_administrator": True,
                        "observer": False
                    },
                    "supports_ldap": False,
                    "enabled": True,
                    "twofactor_auth_domain": None,
                    "attachments_enabled": True,
                    "has_saml": False,
                    "state": None,
                    "addr2": None,
                    "twofactor_cookie_lifetime_secs": 0,
                    "require_saml": False,
                    "tasks_private": False,
                    "authorized_ldap_group": None,
                    "id": 202,
                    "name": "Mock Org"
                }
            ],
            "session_ip": "192.168.56.1",
            "user_fname": "John",
            "user_email": self.email
        }
        cookies = {'JSESSIONID': 'FakeSessionId'}
        return requests_mock.create_response(request,
                                             status_code=200,
                                             cookies=requests.cookies.cookiejar_from_dict(cookies),
                                             json=session_data)

    @resilient_endpoint("GET", "/incidents/[0-9]+$")
    def incident_get(self, request):
        """ Callback for GET to /orgs/<org_id>/incidents/<inc_id> """
        LOG.debug("incident_get")
        return requests_mock.create_response(request,
                                             status_code=200,
                                             json=self.incident)

    @resilient_endpoint("PUT", "/incidents/[0-9]+")
    def incident_put(self, request):
        """ Callback for PUT to /orgs/<org_id>/incidents/<inc_id> """
        LOG.debug("incident_put")
        self.incident = request.json()
        return requests_mock.create_response(request,
                                             status_code=200,
                                             json=self.incident)

    @resilient_endpoint("POST", "/incidents/")
    def incident_post(self, request):
        """ Callback for POST to /orgs/<org_id>/incidents """
        LOG.debug("incident_post")
        self.incident = request.json()
        return requests_mock.create_response(request,
                                             status_code=200,
                                             json=self.incident)

    @resilient_endpoint("GET", "/orgs/[0-9]+$")
    def org_get(self, request):
        """ Callback for GET to /orgs/<org_id> """
        LOG.debug("org_get")
        org_data = {"users": {"2": {"id": 2,
                                    "fname": "Resilient",
                                    "lname": "Sysadmin",
                                    "status": "A",
                                    "email": self.email,
                                    "phone": None,
                                    "cell": None,
                                    "title": None,
                                    "notes": None,
                                    "locked": False,
                                    "enabled": True,
                                    "roles": {
                                        "administrator": False,
                                        "observer": False,
                                        "master_administrator": True,
                                        "create_incs": True
                                    },
                                    "last_login": 1475614308048,
                                    "org_id": 202,
                                    "group_ids": [],
                                    "is_external": False},
                              "3": {"id": 3,
                                    "fname": "User",
                                    "lname": "Three",
                                    "status": "A",
                                    "email": "someone@example.com",
                                    "phone": None,
                                    "cell": None,
                                    "title": None,
                                    "notes": None,
                                    "locked": False,
                                    "enabled": True,
                                    "roles": {
                                        "administrator": False,
                                        "observer": False,
                                        "master_administrator": True,
                                        "create_incs": True},
                                    "last_login": 1476711664869,
                                    "org_id": 202,
                                    "group_ids": [],
                                    "is_external": False}
                              },
                    "actions_framework_enabled": True
                    }
        return requests_mock.create_response(request, status_code=200, json=org_data)

    @resilient_endpoint("GET", "/types/incident/fields")
    def incident_fields_get(self, request):
        """ Callback for GET to /orgs/<org_id>/types/incident/fields """
        LOG.debug("incident_fields_get")
        field_data = test_data("200_JSON_GET__types_incident_fields.json")
        return requests_mock.create_response(request, status_code=200, json=field_data)

    @resilient_endpoint("GET", "/types/actioninvocation/fields")
    def action_fields_get(self, request):
        """ Callback for GET to /orgs/<org_id>/types/actioninvocation/fields """
        LOG.debug("action_fields_get")
        field_data = test_data("200_JSON_GET__types_incident_fields.json")
        return requests_mock.create_response(request, status_code=200, json=field_data)

    @resilient_endpoint("GET", "/types/__function/fields")
    def function_fields_get(self, request):
        """ Callback for GET to /orgs/<org_id>/types/__function/fields """
        LOG.debug("function_fields_get")
        field_data = test_data("200_JSON_GET__types_function_fields.json")
        return requests_mock.create_response(request, status_code=200, json=field_data)

    @resilient_endpoint("GET", "/actions")
    def actions_get(self, request):
        """ Callback for GET to /orgs/<org_id>/actions """
        LOG.debug("actions_get")
        data = test_data("200_JSON_GET__actions.json")
        return requests_mock.create_response(request, status_code=200, json=data)

    @resilient_endpoint("GET", "/message_destinations")
    def message_destinations_get(self, request):
        """ Callback for GET to /orgs/<org_id>/message_destinations """
        LOG.debug("message_destinations_get")
        data = test_data("200_JSON_GET__message_destinations.json")
        return requests_mock.create_response(request, status_code=200, json=data)

    @resilient_endpoint("GET", "/functions")
    def functions_get(self, request):
        """ Callback for GET to /orgs/<org_id>/functions """
        LOG.debug("functions_get")
        data = test_data("200_JSON_GET__functions.json")
        return requests_mock.create_response(request, status_code=200, json=data)

    @resilient_endpoint("GET", "/functions/.*$")
    def functions_get_xxx(self, request):
        """ Callback for GET to /orgs/<org_id>/functions/nnn """
        LOG.debug("functions_get_xxx")
        data = test_data("200_JSON_GET__functions.json")
        return requests_mock.create_response(request, status_code=200, json=data["entities"][0])

    @resilient_endpoint("GET", "/incidents/[0-9]+/attachments/[0-9]+/contents$")
    def attachment_contents_get(self, request):
        """ Callback for GET to attachment contents """
        LOG.debug("attachment_contents_get")
        return requests_mock.create_response(request,
                                             status_code=200,
                                             json='"abcdef"')
